/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Msg.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_MSG_H
#define DCDT_MSG_H

#include <DCDT_Defs.h>
#include <DCDT_Time.h>
#include <DCDT_ctrace.h>
#include <string.h>

// HandShake messages types
/// HandShake Notify 
#define HS_NOTIFY    1
/// HandShake Answer
#define HS_ANSWER    2

// msg types
#define MT_SUBSUPD        1001
#define MT_KEEPALIVE      1002
#define MT_CONNECT        1003

// msg delivery warranties
#define NO_WARRANTY    0
#define MID_WARRANTY   5
#define MAX_WARRANTY  10

#define UDP_WARRANTY          0
#define UDP_RETRY_WARRANTY    5
#define TCP_WARRANTY         10

/// HandShake message header
typedef struct _HSMsgHeader {
  int type;
  int AgoraID;
  int channel;
  int payload_len;
} HSMsgHeader;

#define HSMSGHEADER_LEN sizeof(HSMsgHeader)
#define MAX_HSPAYLOAD_LEN 16

/// Message header
typedef struct _DCDT_MsgHeader {

  /// Type of message
  int type;
  
  /// Member ID of the message creator
  int MemberID;

  /// Agora ID of the message creator
  int AgoraID;

  /// Number of readers for this message
  int Readers_Counter; 
  DCDT_Mutex *mtx;

  DCDT_TIME creation_time;
  int priority;
  int payload_len;
  int delivery_warranty;
} DCDT_MsgHeader;

#define MSGHEADER_LEN sizeof(DCDT_MsgHeader)

/*! \brief Message object 
 *
 *  The payload MUST be dynamically allocated.
 */
class DCDT_Msg {

 public:
  // Constructor with type and priority
  DCDT_Msg(int type, int priority) {
    header = new DCDT_MsgHeader ;
    header->mtx = new DCDT_Mutex;
    payload = NULL;

    SetType( type );

    header->MemberID = 0;
    header->AgoraID = 0;
    header->priority = priority;
    header->creation_time = GetTime();
    header->payload_len = 0;
    header->delivery_warranty = 0;
    header->Readers_Counter = 0;
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Create Msg %p - H:%p P:%p", this, header, payload ));
   };

  // Constructor with type
  DCDT_Msg(int type ) {
    header = new DCDT_MsgHeader ;
    header->mtx = new DCDT_Mutex;
    SetType( type );
    payload = NULL;

    header->MemberID = 0;
    header->AgoraID = 0;
    header->creation_time = GetTime();
    header->priority = 0;
    header->payload_len = 0;
    header->delivery_warranty = 0;
    header->Readers_Counter = 0;
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Create Msg %p - H:%p P:%p", this, header, payload ));
  }

  // Constructor from header
  DCDT_Msg(DCDT_MsgHeader* h) {
    header = h;
    payload = NULL;
    header->Readers_Counter = 0;
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Create Msg %p - H:%p P:%p", this, header, payload ));
 };

  ~DCDT_Msg() {
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Delete Msg %p - H:%p P:%p", this, header, payload ));
    if(payload)
      free(payload);

    if ( header->mtx )
      delete header->mtx;

    delete header;
  };

  inline void SetType(int);
  inline int AllocPayload(size_t size, void *buffer );
  inline int SetPayload(void *p, int len);
  inline int ReadType() const;
  inline int ReadAgoraID() const;
  inline void SetAgoraID(int ID);

  inline int ReadMemberID() const;
  inline void SetMemberID(int ID);

  inline int ReadPriority() const;
  inline int ReadPayloadLen() const;

  inline void * GetPayload() const;
  inline DCDT_MsgHeader * GetHeader() const;

  inline DCDT_TIME ReadCreationTime() const;

  inline void SetDeliveryWarranty( int new_del_warran ); 
  inline int ReadDeliveryWarranty() const;

  // functions to manage Readers_Count 
  inline void Increment_Readers_Counter( bool lock=true );
  inline void Decrement_Readers_Counter( bool lock=true );
  inline int Get_Readers_Counter( bool lock=true ) const;

  // function to lock msg
  inline void Lock();
  inline void Unlock();

 private:

  DCDT_MsgHeader * header;
  void * payload;

};

inline void DCDT_Msg::SetType(int num_type) {
  header->type = num_type;
};

inline int DCDT_Msg::AllocPayload(size_t size, void *buffer ) {
   if ( payload ) {
      free( payload );
   }

   if ( (payload = malloc( size )) ) {
      if ( buffer ) {
         memcpy ( payload, buffer, size );
         header->payload_len = size;
      }

      return SUCCESS;
   }

   return FAILURE;
};

inline int DCDT_Msg::SetPayload(void *p, int len) {
  header->payload_len = len;
  if( payload ) {
    free( payload );
  }
  payload = p;
  return SUCCESS;
};

inline int DCDT_Msg::ReadType() const {
  return ( header->type );
};

inline int DCDT_Msg::ReadAgoraID() const {
  return ( header->AgoraID );
};

inline void DCDT_Msg::SetAgoraID(int ID ){
  header->AgoraID = ID;
};

inline int DCDT_Msg::ReadMemberID() const {
  return( header->MemberID );
};

inline void DCDT_Msg::SetMemberID(int ID){
  header->MemberID = ID;
};

inline int DCDT_Msg::ReadDeliveryWarranty() const {
  return ( header->delivery_warranty );
};

inline void DCDT_Msg::SetDeliveryWarranty(int new_del_warran){
  header->delivery_warranty = new_del_warran;
};

inline int DCDT_Msg::ReadPriority() const {
  return ( header->priority );
};

inline int DCDT_Msg::ReadPayloadLen() const {
  return (header->payload_len);
}

inline void * DCDT_Msg::GetPayload() const {
  return(payload);
};

inline DCDT_MsgHeader * DCDT_Msg::GetHeader() const {
  return( header );
};

inline DCDT_TIME DCDT_Msg::ReadCreationTime() const {
  return( header->creation_time);
};

inline void DCDT_Msg::Increment_Readers_Counter( bool lock )
{
  if ( lock ) {
     header->mtx->lock();
     header->Readers_Counter ++;
     header->mtx->unlock();
  }
  else {
     header->Readers_Counter ++;
  }
}

inline void DCDT_Msg::Decrement_Readers_Counter( bool lock )
{
  if ( lock ) {
     header->mtx->lock();
     header->Readers_Counter --;
     header->mtx->unlock();
  }
  else {
     header->Readers_Counter --;
  }
}

inline int DCDT_Msg::Get_Readers_Counter( bool lock ) const
{
  int tmp;

  if ( lock ) {
     header->mtx->lock();
     tmp = header->Readers_Counter;
     header->mtx->unlock();
  }
  else {
     
  }
  return tmp;
}

inline void DCDT_Msg::Lock()
{
   header->mtx->lock();
}

inline void DCDT_Msg::Unlock()
{
   header->mtx->unlock();
}


#endif //define
