/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_UDPSocket.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef UDPSOCKET_H
#define UDPSOCKET_H

#include <iostream>
#include <netdb.h>
#include <Socket.h>

#include <assert.h>

using namespace std;

class UDPSocket : public Socket
{
 public:
  inline void Open(int fd = 0);
  inline void Send(const DCDT_Msg* msg);
  inline void SendHS(int addr, short port, HSMsgHeader *header, void* payload);
  inline void Send(DCDT_Msg* msg, const char* dest, short port);
  inline void Send(DCDT_Msg* msg, int addr, short port);
  inline DCDT_Msg* Receive();
  inline void ReceiveHS(HSMsgHeader*& header, void*& payload);
  inline void MC_Set(int group, int ifaddr);

 private:
  int multi_addr;
  short multi_port;
};

inline void
UDPSocket::Open(int fd)
{
  if (opened)
    return;
  if (fd) sockfd = fd;
  else if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    throw ChannelError(errno);
  opened = 1;

  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("UDPSocket successfully opened"));
}

static inline int MySendMsg ( int sockfd, msghdr *mh, void *header, ssize_t h_size, void *payload, ssize_t p_size )
{
  iovec iov[1];
  int nsent;

  // Sending header
  iov[0].iov_base = header;
  iov[0].iov_len = h_size;

  mh->msg_iov = iov;
  mh->msg_iovlen = 1;

  errno=0;

  nsent = sendmsg ( sockfd, mh, 0 );
  if ( (nsent < 0 ) ) {
    return ( FAILURE );
  }
  else if ( nsent != h_size ) {
    return ( FAILURE );
  }

  // Sending payload
  if ( p_size != 0 ) {
    iov[0].iov_base = payload;
    iov[0].iov_len = p_size;
    
    mh->msg_iov = iov;
    mh->msg_iovlen = 1;
    
    nsent = sendmsg( sockfd, mh, 0 );
    if ( (nsent < 0 ) ) {
      return ( FAILURE );
    }
    else if ( nsent != p_size ) {
      return ( FAILURE );
    }
  }

  return ( SUCCESS );
}

/*! \brief Send the message \a msg to the remote end of the connection.
 *
 *  \exception TimeOut
 *  \exception ConnError
 */
inline void UDPSocket::Send(const DCDT_Msg* msg)
{
  fd_set wset;
  struct timeval tv, *tv_ptr = NULL;
  msghdr mh;

  if (!opened || !connected)
    return;

  FD_ZERO(&wset);
  FD_SET(sockfd, &wset);
  if (unblock_send_flag)
  {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    tv_ptr = &tv;
  }
  else if (tsend_flag)
  {
    tv.tv_sec = tsend_value.tv_sec;
    tv.tv_usec = tsend_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag)
  {
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, NULL, &wset, NULL, tv_ptr))
  {
    case 0: throw TimeOut();
    case -1: throw ConnError(errno);
  }

  memset(&mh, 0, sizeof(mh));

  if ( MySendMsg ( sockfd, &mh, msg->GetHeader(), MSGHEADER_LEN, msg->GetPayload(), msg->ReadPayloadLen() ) != SUCCESS )
    throw ConnError();
 
  TRC_PRINT ( DCDT_TRC_COMM, TRC0, ("UDP message successfully sent" ));

}


/*! \brief Send an handshake message to \a addr:port ( normally a multicast address )
 *
 *  \exception TimeOut
 *  \exception ConnError
 */
inline void UDPSocket::SendHS(int addr, short port, HSMsgHeader *header, void* payload)
{
  fd_set wset;
  struct timeval tv, *tv_ptr = NULL;
  struct sockaddr_in destaddr;
  msghdr mh;
  iovec iov[2];
  int nsent;

  if (!opened)
    return;

  FD_ZERO(&wset);
  FD_SET(sockfd, &wset);
  if (tsend_flag)
  {
    tv.tv_sec = tsend_value.tv_sec;
    tv.tv_usec = tsend_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag)
  {
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, NULL, &wset, NULL, tv_ptr))
  {
    case 0: throw TimeOut();
    case -1: throw ConnError(errno);
  }

  memset(&destaddr, 0, sizeof(destaddr));
  destaddr.sin_family = AF_INET;
  destaddr.sin_addr.s_addr = (in_addr_t)addr;
  destaddr.sin_port = port;

  memset(&mh, 0, sizeof(mh));
  mh.msg_name = &destaddr;
  mh.msg_namelen = sizeof(destaddr);

  /* Valgrind da' un errore perche' "header" e "payload" possono essere
     non inizializzati ( verificare ) */
  iov[0].iov_base = header;
  iov[0].iov_len = HSMSGHEADER_LEN;
  iov[1].iov_base = payload;
  iov[1].iov_len = header->payload_len;

  mh.msg_iov = iov;
  mh.msg_iovlen = 2;

  if ((nsent = sendmsg(sockfd, &mh, 0)) < 0)
    throw ConnError(errno);
  else if (nsent < (int)(iov[0].iov_len + iov[1].iov_len)) {

    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Error: UDP bytes sent: %i ( Expected %d )", nsent, (int)(iov[0].iov_len + iov[1].iov_len) ));

    throw ConnError();
  }

  /*
  if ( MySendMsg( sockfd, &mh, header, HSMSGHEADER_LEN, payload, header->payload_len ) != SUCCESS )
    throw ConnError();
  */
  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("UDP bytes sent: %i message successfully sent", nsent));

}


/*! \brief Send the message \a msg to the specified address ( \a addr:port ) 
 *
 *  Timers precedence: unblock, send, lost. 
 *  \exception TimeOut
 *  \exception ConnError
 *  \exception ChannelError
 */
inline void UDPSocket::Send(DCDT_Msg* msg, const char* addr, short port)
{
  fd_set wset;
  struct timeval tv, *tv_ptr = NULL;
  struct sockaddr_in destaddr;
  msghdr mh;
  
  if (!opened)
    return;

  FD_ZERO(&wset);
  FD_SET(sockfd, &wset);
  if (unblock_send_flag) {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    tv_ptr = &tv;
  }
  else if (tsend_flag) {
    tv.tv_sec = tsend_value.tv_sec;
    tv.tv_usec = tsend_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag) {
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  
  switch (select(sockfd+1, NULL, &wset, NULL, tv_ptr)) {
  case 0: throw TimeOut();
  case -1: throw ConnError(errno);
  }

  memset(&destaddr, 0, sizeof(destaddr));
  destaddr.sin_family = AF_INET;
  if (inet_pton(AF_INET, addr, &destaddr.sin_addr) < 1) {
    throw ChannelError(errno);
  }
  destaddr.sin_port = port;
  
  memset(&mh, 0, sizeof(mh));
  mh.msg_name = &destaddr;
  mh.msg_namelen = sizeof(destaddr);

  if ( MySendMsg( sockfd, &mh, msg->GetHeader(), MSGHEADER_LEN, msg->GetPayload(), msg->ReadPayloadLen() ) != SUCCESS )
    throw ConnError();
  
  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("UDP message successfully sent" ));  
}

/*! \overload
 */
inline void UDPSocket::Send(DCDT_Msg* msg, int addr, short port)
{
  fd_set wset;
  struct timeval tv, *tv_ptr = NULL;
  struct sockaddr_in destaddr;
  msghdr mh;

  if (!opened)
    return;

  FD_ZERO(&wset);
  FD_SET(sockfd, &wset);
  if (unblock_send_flag)
  {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    tv_ptr = &tv;
  }
  else if (tsend_flag)
  {
    tv.tv_sec = tsend_value.tv_sec;
    tv.tv_usec = tsend_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag)
  {
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, NULL, &wset, NULL, tv_ptr))
  {
    case 0: throw TimeOut();
    case -1: throw ConnError(errno);
  }

  memset(&destaddr, 0, sizeof(destaddr));
  destaddr.sin_family = AF_INET;
  destaddr.sin_addr.s_addr = (in_addr_t)addr;
  destaddr.sin_port = port;

  memset(&mh, 0, sizeof(mh));
  mh.msg_name = &destaddr;
  mh.msg_namelen = sizeof(destaddr);
  
  if ( MySendMsg( sockfd, &mh, msg->GetHeader(), MSGHEADER_LEN, msg->GetPayload(), msg->ReadPayloadLen() ) != SUCCESS )
    throw ConnError();

  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("UDP message successfully sent" ));
}

/*! \brief Receive a message
 *
 *  Timers precedence: unlocking, receive, lost.
 *  \exception ConnError
 *  \return NULL - No message received
 *  \return !=NULL - 
 */
inline DCDT_Msg* UDPSocket::Receive()
{
  DCDT_MsgHeader *msg_header;
  int plen;
  DCDT_Msg *msg;
  void *buf;
  msghdr mh;
  iovec iov[1];
  ssize_t nread;
  fd_set rset;
  struct timeval tv, *tv_ptr = NULL;

  if (!opened)
    return NULL;

  FD_ZERO(&rset);
  FD_SET(sockfd, &rset);
  if (unblock_recv_flag) {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    tv_ptr = &tv;
  }
  else if (trecv_flag) {
    tv.tv_sec = trecv_value.tv_sec;
    tv.tv_usec = trecv_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag) {
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, &rset, NULL, NULL, tv_ptr)) {
  case 0: return NULL;
  case -1: throw ConnError(errno);
  }

  // preparing data structure
  memset(&mh, 0, sizeof(mh));
  msg_header = new DCDT_MsgHeader;
  iov[0].iov_base = msg_header;
  iov[0].iov_len = MSGHEADER_LEN;
  mh.msg_iov = iov;
  mh.msg_iovlen = 1;
  //nread = recvmsg(sockfd, &mh, 0 );
  //msg_header->mtx = new DCDT_Mutex;
  if ((nread = recvmsg(sockfd, &mh, 0)) < 0) {
    delete msg_header;
    throw ConnError(errno);
  }
  else if (nread != MSGHEADER_LEN) {
    delete msg_header;
    //throw ConnError();
    return NULL;
  }
  msg_header->mtx = new DCDT_Mutex;
  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("UDP byte read: %i", nread));

  msg = new DCDT_Msg(msg_header);
  plen = msg->ReadPayloadLen();
  if ( plen != 0 ) {
    buf = malloc ( plen );
    //assert(buf!=NULL);
    if ( buf == NULL ) {
      cout << "UDPSocket: Unable to allocate " << plen << "bytes\n";
      delete msg_header;
      throw ConnError(errno);
    }

    memset ( buf, 0, plen );
    iov[0].iov_base = buf;
    iov[0].iov_len = plen;
    mh.msg_iov = iov;
    mh.msg_iovlen = 1;
    if ( recvmsg(sockfd, &mh, 0) != (ssize_t)plen ) {
      free(buf);
      delete msg_header;
      TRC_PRINT( DCDT_TRC_COMM, TRC1, ("recvmsg error %d", nread));
      throw ConnError(errno);
    }
    
    msg->SetPayload( buf, plen);
  }
  else {
    msg->SetPayload( NULL, 0 );
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("UDP message successfully received"));

  return msg;
}

/*! \brief Receive a system message
 *
 *   Look only at receive timer. The calling procedure must ensure that there is enought room for the received message header and payload.
 *   \exception ConnError
 *   \exception Timeout No message received
 */
inline void UDPSocket::ReceiveHS(HSMsgHeader*& header, void*& payload)
{
  HSMsgHeader *msg_header;
  void *data;
  msghdr mh;
  iovec iov[2];

  int nread = 0;
  fd_set rset;
  struct timeval tv, *tv_ptr = NULL;

  if (!opened)
    return;

  FD_ZERO(&rset);
  FD_SET(sockfd, &rset);

  if (trecv_flag)
  {
    tv.tv_sec = trecv_value.tv_sec;
    tv.tv_usec = trecv_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, &rset, NULL, NULL, tv_ptr)) {
    case 0:
      throw TimeOut();
    case -1:
      throw ConnError(errno);
  }

  // preparing data structure
  cout << "Receive HS\n";
  memset(&mh, 0, sizeof(mh));
  msg_header = new HSMsgHeader();
  data = malloc(MAX_HSPAYLOAD_LEN);
  iov[0].iov_base = msg_header;
  iov[0].iov_len = HSMSGHEADER_LEN;
  iov[1].iov_base = data;
  iov[1].iov_len = MAX_HSPAYLOAD_LEN;
  mh.msg_iov = iov;
  mh.msg_iovlen = 2;
  if ((nread = recvmsg(sockfd, &mh, 0)) < 0) {
    free(data);
    delete msg_header;
    throw ConnError(errno);
  }
  else if (nread == 0 || nread < (int)HSMSGHEADER_LEN) {
    free(data);
    delete msg_header;
    throw ConnError();
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("UDP byte read: %i", nread));

  if (nread < (int)(HSMSGHEADER_LEN + (msg_header->payload_len))) {
    free(data);
    delete msg_header;
    throw ConnError();
  }

  payload = malloc(msg_header->payload_len);
  if ( payload == NULL ) {
     TRC_PRINT( DCDT_TRC_COMM, TRC0, ("payload == NULL requested %d",msg_header->payload_len));
     free(data);
     delete msg_header;
     throw ConnError();
  }

  memcpy(payload, data, msg_header->payload_len);
  free(data);
  header = msg_header;

  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("UDP message successfully received"));
}


/*! \brief Associate us to the multicast group \a group of the interface \a ifaddr.
 *
 *  \exception ChannelError
 */
inline void
UDPSocket::MC_Set(int group, int ifaddr)
{
  struct ip_mreq mreq;

  if (!opened)
    return;
  
  mreq.imr_multiaddr.s_addr = multi_addr = group;
  mreq.imr_interface.s_addr = ifaddr;
  
  if (setsockopt(sockfd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0)
    throw ChannelError(errno);
  
  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Multicast option successfully set"));

  u_char val = (u_char)0;
  
  if (setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_LOOP, &val, sizeof(val)) < 0)
    throw ChannelError(errno);
  
  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Multicast unloop option successfully set"));

}

#endif
