/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**!!!                                                                   
**!!!                   INFO SOLUTION SPA                     
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                    
**!!! PROJECT       : ROBY                                 
**!!!                                                                    
**!!! CLASS         : Message Manager                 
**!!!
**!!! PACKAGE       : DATA_HANDLING
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                   
**!!! FILE NAME     : MessageManager.h
**!!!                                                                   
**!!! CREATION DATE : 18.10.2001         
**!!!                                                                   
**!!! AUTHOR        : P.M.  Roby  Team                             
**!!!                                                                   
**!!!                                                                   
**!!!___________________________________________________________________
**!!!                                                                   
**!!! SPECS ON DOCS : <Spec. Doc Name A>  Rev.: AA.aa                   
**!!!                                                                     
**!!!                 <Spec. Doc Name B>  Rev.: BB.bb                   
**!!!                ............                                       
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**!!!                                                                   
**!!! OVERVIEW      : Message Manager makes to modules transparent the 
**!!!	             sensor message management
**!!!                                                                   
**!!!                                                                   
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
*/

#ifndef __MESSAGE_MANAGER_H
#define __MESSAGE_MANAGER_H
/*----------------------------------------------------------------------
 *--------------------- SYSTEM  INCLUDE FILES --------------------------
 *---------------------------------------------------------------------*/
#include <stl.h>

/*----------------------------------------------------------------------
 *--------------------- PROJECT INCLUDE FILES --------------------------
 *---------------------------------------------------------------------*/
#include <const.h>
#include <msg.h>
#include <stc.h>
#include "ModuleExpert.h"

/*----------------------------------------------------------------------
 *-----------------------  DEFINE MACRO  -------------------------------
 *---------------------------------------------------------------------*/

/*----------------------------------------------------------------------
 *-----------------------  DEFINE TYPES  -------------------------------
 *---------------------------------------------------------------------*/
struct AttributeType
{
  float value;
  float variance;
  float reliability;
  char name[STRING_LENGTH];

  AttributeType(float _value = 0,float _variance = 0,float _reliability = 0, const char* _name = 0)
  {
    value = _value;
    variance = _variance;
    reliability = _reliability;
    if (strlen(_name) < STRING_LENGTH)
      strcpy(name,_name);
  }
};

struct GeneralMsgType
{
  char name[STRING_LENGTH];
  unsigned short kernel_id;
  int sender_id;
  int num_attributes;
  struct timeval timestamp;
  AttributeType attributes[];
};


struct MsgObject
{
  char name[STRING_LENGTH];
  struct timeval timestamp;
  unsigned short kernel_id;
  int sender_id;
  map<const char*,AttributeType,ins_ord> attributes;
  
  void GetValues(GeneralMsgType* obj);
  
  ~MsgObject()
  {
    for_each(attributes.begin(),attributes.end(),destroy_pair_char<AttributeType>());
  }
};

typedef vector<MsgObject*> MsgPayload;

typedef map<unsigned int,MsgPayload,less<int> > GeneralMsgMapType;

struct BrowsingType
{
  unsigned int message;
  unsigned int object;
  unsigned int attribute;

  void Initialize();
  void Set(unsigned int msg,unsigned int obj,unsigned int att);
};

/*----------------------------------------------------------------------
 *-----------------------  DEFINE CLASS  -------------------------------
 *---------------------------------------------------------------------*/

/*! \brief This class is the interface that provided the basic functions for creating and senting messages with attributes ( values with reliability )
 *
 *  The MessageManager class is an abstract class ( since it contains three
 *  pure virtual methods which are inherited from the class @ref ModuleExpert )
 *  that gives the necessary interface to build attribute messages. In order to
 *  write a message, first of all we have to call the method @ref #AddObject
 *  specifying the number that identifies the message type. Then we can fill
 *  the message using the @ref #AddAttribute method. Once all the messages has
 *  been completed, we can send it throught the @ref #SendGeneralMessage or
 *  @ref #SendGeneralMessages.
 *  We can have more than a message in the sending queue at any time.
 */
class MessageManager: public virtual ModuleExpert
{

  /*------------------- DEFINE CONSTRUCTOR/DESTRUCTOR -------------------*/
 protected:
  /** default Constructor **/
  MessageManager(ModuleDispatch* Dispatch);
  MessageManager(ModuleDispatch* Dispatch, DCDT_TIME period);

  /** copy Constructor **/ 
  
  /** Destructor **/
  ~MessageManager(void);
  
  /*---------------------------- PUBLIC ---------------------------*/
 public:
  /** Attribute: **/
  /** Methods: **/

  /*! \brief Send all queued messages of the specified type
   *
   *  After this call the sending queue for the specified type of message is empty.
   *  \param message Type of the messages to send
   */
  void SendGeneralMessage(unsigned int message);

  /*! \brief Send all queued messages
   */
  void SendGeneralMessages();

  /*--------------------------- PROTECTED -------------------------*/
 protected:
  /** Attribute: **/
  /** Methods: **/

  /*! \brief Create a new message of the specified type
   *
   *  \param message Type of the new message
   *  \return identifier of the created message
   */
  unsigned int AddObject(unsigned int message);

  /*! \brief Create a new message of the specified type and with the given name
   *
   *  \param message Type of the message
   *  \param name Message's name
   *  \return identifier of the created message
   */
  unsigned int AddObject(unsigned int message, const char* name);

  /*! \brief Add an attribute to a specified message
   *
   *  Add an attribute to the message identified by type \a message and id \a object.
   *  \param message Type of the message
   *  \param object  Identifier of the message
   *  \param attribute Name for the new attribute
   *  \param value
   *  \param variance 
   *  \param reliability
   */ 
  void AddAttribute(unsigned int message,
		    unsigned int object,
		    const char* attribute, 
		    float value,
		    float variance,
		    float reliability);
  //:TODO: mettere un campo per il tipo di attributo (assoluto,relativo,integrale,...)

  /*! \brief Subscribe to the specified type of messages
   *
   *  \param message Type of the messages that we want to receive
   *  \param request Subscribe to local ( LOCAL ) or remote ( REMOTE ) messages
   */
  void AddGeneralMsgRequest(unsigned int message, RequestType request=LOCAL);

  /*! \brief Unsubscribe from the specified type of messages
   *
   *  \param message Type of the messages that we don't want to receive
   *  \param request Subscribe to local ( LOCAL ) or remote ( REMOTE ) messages
   */
  void RemGeneralMsgRequest(unsigned int message, RequestType request=LOCAL);

  /*! \brief Delete all the messages of the specified type received
   *
   *  \param message Type of the messages to delete
   */
  void EraseGeneralMessage(unsigned int message);

  /*! \brief Delete all the messages received
   */
  void EraseGeneralMessages();

  AttributeType* GetFirstAttribute(unsigned int msg, unsigned int obj);
  AttributeType* GetFirstAttribute(unsigned int msg);
  AttributeType* GetFirstAttribute();
  AttributeType* GetNextAttribute();
  bool IsLastAttributeOfObject();
  bool IsLastAttributeOfMessage();
  bool IsLastAttribute();
  AttributeType* GetAttribute(unsigned int msg, unsigned int obj, unsigned int num);
  vector<AttributeType*> GetAttributeByName(const char* name);
  AttributeType* GetLatestAttributeByName(const char* name);

  unsigned int GetAttributesOfObject(unsigned int msg, unsigned int num);
  int GetSenderOfObject(unsigned int msg, unsigned int num);
  unsigned short GetKernelOfObject(unsigned int msg, unsigned int num);

  unsigned int GetFirstMessage();
  unsigned int GetNextMessage();
  bool IsLastMessage();
  bool IsArrivedMessage(unsigned int num);
  unsigned int GetMessage(unsigned int pos);
  unsigned int GetObjectsOfMessage(unsigned int num);

  MsgObject* GetFirstObject();
  MsgObject* GetNextObject();
  MsgObject* GetCurrentObject();
  map<const char*,AttributeType,ins_ord>* GetAttributesOfCurrentObject();
  void RemoveCurrentObject();
  vector<MsgObject*>* GetAllObjectsOfMessage(unsigned int message);

  void ReceiveGeneralMessage(unsigned short message);

  void ReceiveGeneralMessages();
  
  virtual void RunInit()=0;
  virtual void RunDuty()=0;
  virtual void RunClose()=0;

  /*--------------------------- PRIVATE ---------------------------*/
 private:
  /** Attribute: **/
  /// List with all the messages requested
  GeneralMsgMapType mRequestedMessages;

  /// List with all the messages that we have to sent
  GeneralMsgMapType mSendingMessages;
  BrowsingType mBrowsingInfo;

  /** Methods: **/
  int SizeOfMessage(GeneralMsgType* message);
  int SizeOfMessage(MsgObject* message);
};
/*------------------- INLINE IMPLEMENTATION ------------------------*/

#endif
/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*-------------------------- End Of Header File --------------------*/
/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
