/*
CONTROLLER.C
Controller module
This module contains the control algorithms for both motors. The control laws could be the same for both motors, but they could be controlled in different way (one motor controlled in speed, the other controlled in position).
It is possible every type of control: PID control, fuzzy control, etc. If it is necessary to change control law, we must modify this file.

Now we use an incremental PID Controller. Control law is:

u(k) = u(k-1) + Kp*[e(k) - e(k-1)] + Ki*e(k) + Kd*[y(k)-y(k-1) - y(k-1)-y(k-2)]

*/

#include <stdlib.h>
#include <timers.h>
#include <pwm.h>
#include <string.h>

#include "definitions.h"
#include "controller.h"
#include "rs232.h"
#include "eeprom.h" // aggiunto per matching


int FuzzyOut=0;

int tick_0=0, tick_1=0;
	
int y_0, y_1; // measured process output, k th step
static int y_0_k_1 = 0, y_1_k_1 = 0;   // k-1 th step
static int y_0_k_2 = 0, y_1_k_2 = 0;   // k-2 th step

int e_0, e_1; // current error
static int e_0_k_1 = 0, e_1_k_1 = 0; // k-1 th step error

int u_0, u_1; // control action, k th step
static int u_0_k_1 = 0, u_1_k_1 = 0; // k-1 th step
	
static int setpoint_0_k_1 = 0, setpoint_1_k_1 = 0; //k-1 th setpoints
	
static unsigned char dir_0=0, dir_1=0; //engine 0 and engine 1 directions

unsigned char rules[256-F_RULES_ADDR];
int rules_conseguent[256-F_RULES_ADDR];
unsigned char rules_matching[256-F_RULES_ADDR];	

	unsigned char error_in_mf[5];
	unsigned char changerr_in_mf[7];

unsigned char lastIEaddress;
unsigned char lastIEvalue;

int matching_from_table(int error, int mf, int input) {
	int mf_sign = +1;
	int first_address = MF1_B_ADDR; 
	char value_from_table;
	unsigned char matched_value = 0;
	unsigned char temp_mf = 0;
	unsigned char cacheIEaddress;

	if( input == 2 ) // sto valutando l'errore
		first_address = MF2_B_ADDR;
	else if (input != 1); // c' stato un errore
			
	// limito l'error da -100 a 100
	if( error<-100 ) error = -100;
	else {if( error>100 ) error = 100;}

	if( error<0 ) {
		error = -error;
		mf_sign = -1;
	}
	// a questo punto error  per forza compreso tra
    // 0 e 100 estremi compresi
	cacheIEaddress=error+first_address;
	if(cacheIEaddress == lastIEaddress){
		value_from_table = lastIEvalue;
	}else{
		value_from_table = readByteFromIE((unsigned char)(cacheIEaddress));
		lastIEvalue=value_from_table;
		lastIEaddress=cacheIEaddress;
	}
	/* I 6 bit meno significativi costituiscono il valore di matching */
	matched_value = value_from_table & (0xFF>>(2));
	/* I 2 bit pi significativi costituiscono le due mf */
	temp_mf = temp_mf | (value_from_table & (0xFF<<(6)));

	// il grado di match si riferisce sempre alla mf pi bassa!!
	switch (temp_mf) {
   		case 0: 
			// mf 0 e 1
			if( mf == 0 ) return matched_value;
			if( mf*mf_sign == 1 ) return 63 - matched_value;
			return 0;
		break; // non raggiungibile
		case 1: // 1
			// mf 0 e 3: caso impossibile!!!
		break;
		case 128: // 2
			// mf 1 e 2
			if( mf*mf_sign == 1 ) return matched_value;
			if( mf*mf_sign == 2 ) return 63 - matched_value;
			return 0;
		break; // non raggiungibile
		case 192: // 3
			// mf 2 e 3
			if( mf*mf_sign == 2 ) return matched_value;
			if( mf*mf_sign == 3 ) return 63 - matched_value;
			return 0; // non raggiungibile
		break;
		default: // error
			return -1;
	}
}

int FuzzyController(int error, int change_error){
	// inputs are error and change_error
	// output is a voltage increment
		int i =0;
		int mirror_sign=1;
		int numeratore = 0, denominatore =0;
		int temp_rul_1=0,temp_rul_2=0;

		error=(error*100)/PWM_MAX;
		change_error=(change_error*100)/PWM_MAX;


		/* Fuzzyfication */
		for(i=0;i<5;i++){
			error_in_mf[i] = (unsigned char) matching_from_table(error, i-2, 1);
		}

		for(i=0;i<7;i++){
			changerr_in_mf[i] = (unsigned char) matching_from_table(change_error, i-3, 2);
		}

		mirror_sign=1;
		if(error<0)mirror_sign=-1;


		/* Rules decoding and weighting */  
		for(i=0;i<num_of_rules;i++){

	/*The MF of first input variable (error) is codified by the 2 most significative bit
		11 none (dont care) */

			if( ((rules[i])>>6) == 0x03 ){
				temp_rul_1=7;
			}else{
				temp_rul_1= ( ((int)(rules[i])>>6 ) );
				if(mirror_sign<0){temp_rul_1 = (-temp_rul_1)+2;}else{temp_rul_1 = temp_rul_1 +2;}
			}  //0-1-2-3-4  7=none

	/*The MF of second input variable (change_error) is codified by 3 bit
		111 none (dont care) */

			if ( ((rules[i]&0x38)>>3) == 0x07 ){
				temp_rul_2=7;
			}else{
				temp_rul_2=( ((int)((rules[i]&0x38)>>3)) -3); 
				if(mirror_sign<0){temp_rul_2 = (-temp_rul_2)+3; }else{temp_rul_2 = temp_rul_2 +3;}
			}  //0-1-2-3-4-5-6   7=none

			if( temp_rul_1==7 && temp_rul_2!=7 ) {
				rules_matching[i]= changerr_in_mf[temp_rul_2];
			}else{
				if( temp_rul_2==7 && temp_rul_1!=7 ) {
					rules_matching[i]= error_in_mf[temp_rul_1];
				}else{
					if( changerr_in_mf[temp_rul_2] < error_in_mf[temp_rul_1] ) rules_matching[i]= changerr_in_mf[temp_rul_2];
					else rules_matching[i]= error_in_mf[temp_rul_1];
				}
			}

			if(rules_matching[i]>0){
				numeratore = numeratore + (((int)rules_matching[i])*rules_conseguent[i]);
				denominatore = denominatore + ((int)rules_matching[i]); 
			}
		}

		if (denominatore<=0) return 0;	
		FuzzyOut= numeratore/denominatore;
		if(f_incr_max!=100) FuzzyOut=(FuzzyOut*((int)f_incr_max))/100;
		if(mirror_sign<0)FuzzyOut= -FuzzyOut;
		return FuzzyOut;
}

//engine controller subroutine
void engine_control() {
	static unsigned int debug_counter = 0; // used to print debug messages
	int f=0;

	//read engine direction
	dir_0 = DIR_ENGINE_0;
	dir_1 = DIR_ENGINE_1;
	//read tick from encoder
	tick_0 = ReadTimer0();
	tick_1 = ReadTimer1();
	//reset encoders
	WriteTimer0(0);
	WriteTimer1(0);
	
	if (!tuning) {	
		//here we are not in auto tuning mode
		/*in normal mode (not in auto-tuning) the loop is closed and the control is enabled*/

		//if directions is backward, I must invert tick value (se va il bit di direzione)
		if (dir_0 == ENGINE_BACKWARD) tick_0 = -tick_0;
		if (dir_1 == ENGINE_BACKWARD) tick_1 = -tick_1;
		//if (setpoint_0 < 0) tick_0 = -tick_0;		
		//if (setpoint_1 < 0) tick_1 = -tick_1;		

		//update total tick counters
		//total tick for motor0
		total_tick_0 += tick_0;
		//total tick for motor1
		total_tick_1 += tick_1;
	
		if (already_read_ticks_0) {
			tick_0 += already_read_ticks_0;
			already_read_ticks_0 = 0;
		}
		if (already_read_ticks_1) {
			tick_1 += already_read_ticks_1;
			already_read_ticks_1 = 0;
		}
	
		//calculate the controller output
		y_0 = (int) tick_0 * ticks_to_pwm_0;
		y_1 = (int) tick_1 * ticks_to_pwm_1;
		
		//calculate errors
		e_0 = setpoint_0 - y_0;
		if (setpoint_0_k_1 != setpoint_0) {
			//if the setpoint has changed, we reset the integral error
	//		u_0_k_1 = 0; tolta per evitare scattino al cambio setpoint
			e_0_k_1 = 0;
		}
		
		e_1 = setpoint_1 - y_1;
		if (setpoint_1_k_1 != setpoint_1) {
			//if the setpoint has changed, we reset the integral error
	//		u_1_k_1 = 0; tolta per evitare scattino al cambio setpoint
			e_1_k_1 = 0;
		}
		

		//********************************************************
			
		if ( is_fuzzy == 1 ){
		// inputs are error and change_error
		// INVARIANTI error ha 5 MFs; change_in_error ha 3 MFs
		// INVARIANTI l regole sono quelle che sono: codificate sotto
		// Possibili cambiamenti da parte dell'utente: forme MFs e quindi tabelle; pesi delle regole

	
		u_0 = u_0_k_1 + FuzzyController(e_0,e_0-e_0_k_1);
		u_1 = u_1_k_1 + FuzzyController(e_1,e_1-e_1_k_1);

		} else {

			/*
			Here there are the control laws for the motor0 and motor1.
			If we want to change the control laws, we must modify here.
			Now the control law for both motor0 and motor1 is
			u(k) = u(k-1) + Kp*[e(k) - e(k-1)] + Ki*e(k) + Kd*[y(k)-y(k-1) - y(k-1)-y(k-2)]
			*/	
		
			//controll action for motors with pid controller
			u_0 = u_0_k_1 + kp_0*( e_0 - e_0_k_1) + ki_0*e_0 + kd_0*( y_0 - 2*y_0_k_1 + y_0_k_2);
			u_1 = u_1_k_1 + kp_1*( e_1 - e_1_k_1) + ki_1*e_1 + kd_1*( y_1 - 2*y_1_k_1 + y_1_k_2);

		}	
		//**********************************************************
			
		// saturation
		if(u_0 > PWM_MAX) 
			u_0 = PWM_MAX;
		else if (u_0 < PWM_MIN)
			u_0 = PWM_MIN;
	
		if(u_1 > PWM_MAX) 
			u_1 = PWM_MAX;
		else if (u_1 < PWM_MIN)
			u_1 = PWM_MIN;
		
		//set new duty cicle
		SetDCPWM1(PWM_ZERO + u_0);		
		SetDCPWM2(PWM_ZERO + u_1);
	
		// update y
		y_0_k_2 = y_0_k_1;
		y_0_k_1 = y_0;
	
		y_1_k_2 = y_1_k_1;
		y_1_k_1 = y_1;			
	
		// update errors
		e_0_k_1 = e_0;
		e_1_k_1 = e_1;
	
		// update output
		u_0_k_1 = u_0;
		u_1_k_1 = u_1;
		
		//update setpoints
		setpoint_0_k_1 = setpoint_0;
		setpoint_1_k_1 = setpoint_1;
		
		// update motor speeds
		current_speed_0 = tick_0 * ticks_to_speed_0;
		current_speed_1 = tick_1 * ticks_to_speed_1;
	}
	
	// tuning instructions
	//if (tuning) 
	else {
		//we are in auto tuning mode
		/* in auto-tuning mode the control must be disabled because to reach the max speed we must work in open-loop*/
		if (tick_0 > InputData0)
			InputData0 = tick_0;
		if (tick_1 > InputData1)
			InputData1 = tick_1;
		tuning--;
	}

	// DEBUG INFORMATIONS
	if (debug >= DEBUG_FULL && debug_counter++ >= 50) {
		//print these information only if debug level is FULL
		char string_temp[10];
		
		//print setpoint values
		putrsUSART("Sp0: ");
		putsUSART(itoa(setpoint_0, string_temp));
		putrsUSART(" ; Sp1: ");
		putsUSART(itoa(setpoint_1, string_temp));
		WriteUSART(' ');
		
		//print direction values
		putrsUSART("D0: ");
		putsUSART(itoa(dir_0, string_temp));
		putrsUSART(" ; D1: ");
		putsUSART(itoa(dir_1, string_temp));
		WriteUSART(' ');
		
		//print encoder values
		putrsUSART("En0: ");
		putsUSART(itoa(y_0, string_temp));
		putrsUSART(" ; En1: ");
		putsUSART(itoa(y_1, string_temp));
		WriteUSART(' ');
				
		//print tick values
		putrsUSART("Tk0: ");
		putsUSART(itoa(tick_0, string_temp));
		putrsUSART(" ; Tk1: ");
		putsUSART(itoa(tick_1, string_temp));	
		WriteUSART(' ');
		
		//print pwm values
		putrsUSART("P0: ");
		putsUSART(itoa(u_0, string_temp));
		putrsUSART(" ; P1: ");
		putsUSART(itoa(u_1, string_temp));		
		WriteUSART('\r');
		
		//reset debug counter
		debug_counter=0;
	}
}

void error_to_zero(void){
	setpoint_0 = 0;
	setpoint_1 = 0;
	u_0_k_1 = 0; 
	u_0 = 0;
	u_1_k_1 = 0;
	u_1 = 0;
	y_0 = 0;
	y_1 = 0;
	y_0_k_1 = 0; 
	y_1_k_1 = 0;
	y_0_k_2 = 0;
	y_1_k_2 = 0;
	e_0=0;
	e_1=0;
	e_0_k_1=0;
	e_1_k_1=0;
}
