/*
 * AIRboardUI: graphical interface program to interact with airboard
 * through RS232 serial port.
 *
 * In order to compile this package you have to install the
 * Java Communications API on your system.
 * http://java.sun.com/products/javacomm/downloads/index.html
 * 
 * Bruschieri Stefano
 * Campari PierPaolo
 */

package airboardUI;
import javax.comm.*;
import java.util.Vector;
import airboardUI.menu.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;
import java.io.*; // TEST!!!

/**
 * Creates the main window of <code>AIRboardUI</code>. 
*/

public class AirboardFrame extends Frame implements ActionListener {

    final int HEIGHT = 450;
    final int WIDTH  = 410;
    
    /**
     * DELAY_TIME is the number of milliseconds we have to wait
     * before transferring a new command via serial port
     * */
    public static int DELAY_TIME = 1000; 
    
    /**
     * FIRST_INPUT_FILE is the name of file containing the first input array
     * of fuzzy data.
     */
    public static String FIRST_INPUT_FILE = "airboardUI/data/first_input.dat";
    
    /**
     * SECOND_INPUT_FILE is the name of file containing the second input array
     * of fuzzy data.
     */
    public static String SECOND_INPUT_FILE = "airboardUI/data/second_input.dat";
    
    /**
     * RULES_FILE is the name of file containing the fuzzy rules used by the class
     * Rules. Rules must be in decimal format.
     */
    public static String RULES_FILE = "airboardUI/data/rules.dat";
    
    private MenuBar mb;
    private Menu fileMenu;
    private Menu fuzzyMenu;
    private Menu motorsMenu;
    private Menu tuningMenu;
    private Menu enableMenu;
    private Menu disableMenu;
    private Menu aboutMenu;
    private MenuItem mfItem;
    private MenuItem fuzzyEnableItem;
    private MenuItem rulesItem;
    private MenuItem fuzzyFormItem;
    private MenuItem pidFormItem;
    private MenuItem tuningMotOneItem;
    private MenuItem enableMotOneItem;
    private MenuItem disableMotOneItem;
    private MenuItem aboutItem;
    
    private Button openButton;
    private Button closeButton;
    private Button breakButton;
    private Panel buttonPanel;

    private Panel    messagePanel;
    private TextArea messageAreaOut;
    private TextArea messageAreaIn;

    private ShowSettingsPanel showSettingsPanel;
    private SerialParameters parameters;
    private SerialConnection connection;

    private Properties props = null;

    /**
    * Create new <code>AIRboardUI</code> and initilizes it. Parses args to
    * find configuration file. If found, initial state it set to parameters
    * in configuration file.
    *
    * @param args command line arguments used when program was invoked.
    */
    public AirboardFrame(String[] args){
	super("AIRboard UI");

	parameters = new SerialParameters();

	// Set up the GUI for the program
	addWindowListener(new CloseHandler(this));

	mb = new MenuBar();

	// Create the file menu
	fileMenu = new FileMenu(this);
	mb.add(fileMenu);

	motorsMenu = new Menu("Motors");
	tuningMenu = new Menu("Tuning");
	
	tuningMotOneItem =new MenuItem("Tuning mot 1");
	tuningMotOneItem.addActionListener(this);
	tuningMenu.add(tuningMotOneItem);
	
	tuningMenu.add(new MenuItem("Tuning mot 2"));
	tuningMenu.add(new MenuItem("Tuning both mot"));
	motorsMenu.add(tuningMenu);
	
	enableMenu = new Menu("Enable");
	
	enableMotOneItem =new MenuItem("Enable motor 1");
	enableMotOneItem.addActionListener(this);
	enableMenu.add(enableMotOneItem);

	enableMenu.add(new MenuItem("Enable motor 2"));
	enableMenu.add(new MenuItem("Enable both motors"));
	motorsMenu.add(enableMenu);
	
	disableMenu = new Menu("Disable");
	
	disableMotOneItem =new MenuItem("Disable motor 1");
	disableMotOneItem.addActionListener(this);
	disableMenu.add(disableMotOneItem);

	disableMenu.add(new MenuItem("Disable motor 2"));
	disableMenu.add(new MenuItem("Disable both motors"));
	motorsMenu.add(disableMenu);
	mb.add(motorsMenu);
	
	fuzzyMenu = new Menu("Fuzzy");
	
	mfItem = new MenuItem("Copy table into IE");
	mfItem.addActionListener(this);
	fuzzyMenu.add(mfItem);
	
	rulesItem = new MenuItem("Copy rules into IE");
	rulesItem.addActionListener(this);
	fuzzyMenu.add(rulesItem);
	
	fuzzyEnableItem = new MenuItem("Use fuzzy controller");
	fuzzyEnableItem.addActionListener(this);
	fuzzyMenu.add(fuzzyEnableItem);
	
	// TEST!!!!
	MenuItem testItem = new MenuItem("Read kd0");
	testItem.addActionListener(this);
	fuzzyMenu.add(testItem);
	
	fuzzyFormItem = new MenuItem("Insert fuzzy values...");
	fuzzyFormItem.addActionListener(this);
	fuzzyMenu.add(fuzzyFormItem);
	
	mb.add(fuzzyMenu);
	
	PIDMenu pidMenu = new PIDMenu(this);	
	mb.add(pidMenu);
	
	aboutMenu = new Menu("About");
	aboutItem = new MenuItem("About");
	aboutItem.addActionListener(this);
	aboutMenu.add(aboutItem);
	mb.add(aboutMenu);
	setMenuBar(mb);

	messagePanel = new Panel();
	messagePanel.setLayout(new GridLayout(2, 1));

	messageAreaOut = new TextArea();
	messagePanel.add(messageAreaOut);

	messageAreaIn = new TextArea();
	messageAreaIn.setEditable(false);
	messagePanel.add(messageAreaIn);

	add(messagePanel, "Center");

	showSettingsPanel = new ShowSettingsPanel(this);
	
	buttonPanel = new Panel();

	openButton = new Button("Open Port");
	openButton.addActionListener(this);
	buttonPanel.add(openButton);

	closeButton = new Button("Close Port");
	closeButton.addActionListener(this);
	closeButton.setEnabled(false);
	buttonPanel.add(closeButton);

	breakButton = new Button("Send Break");
	breakButton.addActionListener(this);
	breakButton.setEnabled(false);
	buttonPanel.add(breakButton);
	
	Panel southPanel = new Panel();

	GridBagLayout gridBag = new GridBagLayout();
	GridBagConstraints cons = new GridBagConstraints();

	southPanel.setLayout(gridBag);

	cons.gridwidth = GridBagConstraints.REMAINDER;
	gridBag.setConstraints(showSettingsPanel, cons);
	cons.weightx = 1.0;
	southPanel.add(showSettingsPanel);
	gridBag.setConstraints(buttonPanel, cons);
	southPanel.add(buttonPanel);

	add(southPanel, "South");

	parseArgs(args);

	connection = new SerialConnection(this, parameters, 
					  messageAreaOut, messageAreaIn);
	//System.out.println("okko" +parameters.getPortName());
	showSettingsPanel.updateSettingsPanel();

	Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();

	setLocation(screenSize.width/2 - WIDTH/2, 
		    screenSize.height/2 - HEIGHT/2);

	setSize(WIDTH, HEIGHT);
    }

    /**
     * Loads data from file with the specified name. Processed files must
     * be a sequence of numbers separated by '\n'; this method read the numbers,
     * casts them to Integer and puts them in a vector.
     * 
     * @param fileReference File from which this method loads data.
     * @return A vector of Integer read from file passed as argument.
     * @throws IOException
     */
    public Vector loadFuzzyData(String fileReference) throws IOException {
    	Vector toReturn = new Vector();
    	File first_inputFile = null;
    	FileInputStream first_in = null;
		
    	// Load data from files
        first_inputFile = new File(fileReference);
        first_in = new FileInputStream(first_inputFile);
        
        int first_c, second_c, i=0, init=0;
        byte[] cbuf = new byte[(int)first_inputFile.length()];
        byte[] temp;
            	
        first_c = first_in.read(cbuf);
        
        for(int j=0; j<(int)first_inputFile.length(); j++){
        	if( cbuf[j] == '\n'){ 
        		temp = new byte[(j-init)];
        		for(int k=0; k<(j-init-1); k++){
        			temp[k] = cbuf[init+k];
        		}
        		String stringa = new String(temp);
        		toReturn.add(Integer.valueOf(stringa.substring(0,(stringa.length()-2))));
        		i++;
        		init = (j+1);
        	}
        }
        first_in.close();
    	return toReturn;
    }
    
    /**
     * Gets the serial connection.
     *
     * @return The serial connection used by <code>AIRboardUI</code>.
     */
    public SerialConnection getConnection(){
    	return this.connection;	
    }
        
    /**
     * Gets the serial parameters.
     *
     * @return The serial parameters.
     */
    public SerialParameters getParameters(){
    	return this.parameters;	
    }
    
    /**
    * Sets the GUI elements on the configurationPanel.
    */
    public void updateSettingsPanel() {
    	showSettingsPanel.updateSettingsPanel();
    }

    /**
    * Responds to the menu items and buttons.
    */
    public void actionPerformed(ActionEvent e) {
	String cmd = e.getActionCommand();

	// TEST!!!
	if(cmd.equals("Read kd0")){
		SerialPort sPort = connection.getSerialPort();
		OutputStream os;
		String request = "erkd0";
    	try {
    	    os = sPort.getOutputStream();
    	    for(int j=0; j<request.length(); j++){
		    	os.write((int)request.charAt(j));
    	    } 
    	    os.write((int)'\n');
    	} catch (IOException ioe) {
    	    sPort.close();
			System.err.println("OutputStream write error: " + ioe);
    	}
	}
	
	//Disables first motor
	if(cmd.equals("Disable motor 1")){
		if( connection.getSerialPort() == null ){
			AlertDialog np = new AlertDialog(this, "No port is available",
					"In order to transmit any data"+
					"you must"+
					"activate a port.");
		} else {
			this.connection.sendAString("b0");
		}
	}
	
	//Enables first motor
	if(cmd.equals("Enable motor 1")){
		if( connection.getSerialPort() == null ){
			AlertDialog np = new AlertDialog(this, "No port is available",
					"In order to transmit any data"+
					"you must"+
					"activate a port.");
		} else {
			//System.out.println("a0"); //DEBUG
			this.connection.sendAString("a0");
		}
	}
	
	//Tunes first motor
	if(cmd.equals("Tuning mot 1")){
		if( connection.getSerialPort() == null ){
			AlertDialog np = new AlertDialog(this, "No port is available",
					"In order to transmit any data"+
					"you must"+
					"activate a port.");
		} else {
			System.out.println("t0"); //DEBUG
			this.connection.sendAString("t0");
		}
	}

	//Insert fuzzy values: incr max, small out and high out
	if(cmd.equals("Insert fuzzy values...")){
		if( connection.getSerialPort() == null ){
			AlertDialog np = new AlertDialog(this, "No port is available",
					"In order to transmit any data"+
					"you must"+
					"activate a port.");
		} else {
				FuzzyForm ff = new FuzzyForm(this.connection, this);
		}
	}
	
	// Set controller type to fuzzy one 
	if(cmd.equals("Use fuzzy controller")){
		if( connection.getSerialPort() == null ){
			AlertDialog np = new AlertDialog(this, "No port is available",
					"In order to transmit any data"+
					"you must"+
					"activate a port.");
		} else {
			SerialPort sPort = connection.getSerialPort();
			OutputStream os;
			String request = "ewfisfuzzy=+001&+000";
			try {
				os = sPort.getOutputStream();
				for(int j=0; j<request.length(); j++){
					os.write((int)request.charAt(j));
				} 
				os.write((int)'\n');
			} catch (IOException ioe) {
				sPort.close();
				System.err.println("OutputStream write error: " + ioe);
			}
		}
	}

	// Copy fuzzy rules into Internal EEPROM
	if(cmd.equals("Copy rules into IE")){
		if( connection.getSerialPort() == null ){
			AlertDialog np = new AlertDialog(this, "No port is available",
					"In order to transmit any data"+
					"you must"+
					"activate a port.");
		} else {
			try {
				Rules rules = new Rules(this /*, connection*/);
				rules.transmitRules();
			} catch(IOException ioe){
				AlertDialog fe = new AlertDialog(this, "File opening error!",
						"Files containing fuzzy data"+ 
						"don't exists."+
						ioe.toString());
			}
		}
	}	
	
	
	// Copy fuzzy table into Internal EEPROM
	if(cmd.equals("Copy table into IE")){
		//System.out.println(connection);
		if( connection.getSerialPort() == null ){
			AlertDialog np = new AlertDialog(this, "No port is available",
					"In order to transmit any data"+
					"you must"+
					"activate a port.");
		} else {
			try {
				FuzzyTable ft = new FuzzyTable(this, connection);
				ft.transmitTable();
			} catch(IOException ioe){
				AlertDialog fe = new AlertDialog(this, "File opening error!",
						"Files containing fuzzy data"+ 
						"don't exists."+
						ioe.toString());
				//System.out.println("IOexception");
			}
		}
	}
	
	// Opens the about window
	if (cmd.equals("About")) {
	    AboutFrame af = new AboutFrame();
	}

	// Opens a port.
	if (cmd.equals("Open Port")) {
	    openButton.setEnabled(false);
	    Cursor previousCursor = getCursor();
	    setNewCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

	    try {
	    	connection.openConnection();
	    } catch (SerialConnectionException e2) {
		AlertDialog ad = new AlertDialog(this, 
					 "Error Opening Port!",
					 "Error opening port,"+
					 e2.getMessage() + "."+
					 "Select new settings, try again.");
	        openButton.setEnabled(true);
		setNewCursor(previousCursor);
		return;
	    }
	    System.out.println(this.parameters);
	    portOpened();
	    setNewCursor(previousCursor);
	}

	// Closes a port.
	if (cmd.equals("Close Port")) {
	    portClosed();
	}

	// Sends a break signal to the port.
	if (cmd.equals("Send Break")) {
	    connection.sendBreak();
	}
    }

    /**
    * Toggles the buttons to an open port state.
    */
    public void portOpened() {
    	openButton.setEnabled(false);
    	closeButton.setEnabled(true);
    	breakButton.setEnabled(true);
    }

    /**
    * Calls closeConnection on the SerialConnection and toggles the buttons
    * to a closed port state.
    */
    public void portClosed() {
    	connection.closeConnection();
    	openButton.setEnabled(true);
    	closeButton.setEnabled(false);
    	breakButton.setEnabled(false);
    }

    /**
    * Sets the <code>Cursor</code> for the application.
    * @param c New <code>Cursor</code>
    */
    private void setNewCursor(Cursor c) {
    	setCursor(c);
    	messageAreaIn.setCursor(c);
    	messageAreaOut.setCursor(c);
    }

    /**
    * Cleanly shuts down the applicaion. first closes any open ports and
    * cleans up, then exits.
    */
    public void shutdown() {
    	connection.closeConnection();
    	System.exit(1);
    }

    /**
    * Finds configuration file in arguments and creates a properties object from
    * that file.
    */
    private void parseArgs(String[] args) {
	if (args.length < 1) {
	    return;
	}

	File f = new File(args[0]);

	if (!f.exists()) {
	    f = new File(System.getProperty("user.dir") 
			 + System.getProperty("path.separator") 
			 + args[0]);
	}

	if (f.exists()) {
	    try {
		FileInputStream fis = new FileInputStream(f);
		props = new Properties();
		props.load(fis);
		fis.close();
		loadParams();
	    } catch (IOException e) {
	    }
	}
    }

    /**
    * Set the parameters object to the settings in the properties object.
    */
    private void loadParams() {
    	parameters.setPortName(props.getProperty("portName"));
    	parameters.setBaudRate(props.getProperty("baudRate"));
    	parameters.setFlowControlIn(props.getProperty("flowControlIn"));
    	parameters.setFlowControlOut(props.getProperty("flowControlOut"));
    	parameters.setParity(props.getProperty("parity"));
    	parameters.setDatabits(props.getProperty("databits"));
    	parameters.setStopbits(props.getProperty("stopbits"));
	
    	updateSettingsPanel();
    }
 
     

    /**
    * GUI element that holds the user changable elements for connection
    * configuration.
    */
    class ShowSettingsPanel extends Panel {

	private Frame parent;

	private Label portNameLabel;
	private Label showPortNameLabel;

	private Label baudLabel;
	private Label showBaudLabel;

	private Label flowControlInLabel;
	private Label showFlowControlInLabel;

	private Label flowControlOutLabel;
	private Label showFlowControlOutLabel;

	private Label databitsLabel;
    private Label showDatabitsLabel;

	private Label stopbitsLabel;
    private Label showStopbitsLabel;

	private Label parityLabel;
    private Label showParityLabel;

	/**
	Creates and initilizes the configuration panel. The initial settings
	are from the parameters object.
	*/
	public ShowSettingsPanel(Frame parent) {
	    this.parent = parent;

	    setLayout(new GridLayout(4, 4));

	    portNameLabel = new Label("Port Name:", Label.LEFT);
	    add(portNameLabel);
	    showPortNameLabel = new Label(parameters.getPortName(), Label.RIGHT);
	    add(showPortNameLabel);
	    
	    baudLabel = new Label("Baud Rate:", Label.LEFT);
	    add(baudLabel);
	    showBaudLabel = new Label("Baud", Label.RIGHT);
	    add(showBaudLabel);

	    flowControlInLabel = new Label("Flow Control In:", Label.LEFT);
	    add(flowControlInLabel);
	    showFlowControlInLabel = new Label(parameters.getFlowControlInString(), Label.RIGHT);
	    add(showFlowControlInLabel);

	    flowControlOutLabel = new Label("Flow Control Out:", Label.LEFT);
	    add(flowControlOutLabel);
	    showFlowControlOutLabel = new Label(parameters.getFlowControlOutString(), Label.RIGHT);
	    add(showFlowControlOutLabel);

	    databitsLabel = new Label("Data Bits:", Label.LEFT);
	    add(databitsLabel);
	    showDatabitsLabel = new Label(parameters.getDatabitsString(), Label.RIGHT);
	    add(showDatabitsLabel);

	    stopbitsLabel = new Label("Stop Bits:", Label.LEFT);
	    add(stopbitsLabel);
	    showStopbitsLabel = new Label(parameters.getStopbitsString(), Label.RIGHT);
	    add(showStopbitsLabel);

	    parityLabel = new Label("Parity:", Label.LEFT);
	    add(parityLabel);
	    showParityLabel = new Label(parameters.getParityString(), Label.RIGHT);
	    add(showParityLabel);
	}

	/**
	Sets the configuration panel to the settings in the parameters object.
	*/
	public void updateSettingsPanel() {
		showPortNameLabel.setText(parameters.getPortName());
	    showBaudLabel.setText(parameters.getBaudRateString());
	    showFlowControlInLabel.setText(parameters.getFlowControlInString());
	    showFlowControlOutLabel.setText(parameters.getFlowControlOutString());
	    showDatabitsLabel.setText(parameters.getDatabitsString());
	    showStopbitsLabel.setText(parameters.getStopbitsString());
	    showParityLabel.setText(parameters.getParityString()); 
	}

	/**
	Sets the parameters object to the settings in the configuration panel.
	*/
	/*
        public void setParameters() {
        	
            parameters.setPortName(portChoice.getSelectedItem());
	    parameters.setBaudRate(baudChoice.getSelectedItem());
            parameters.setFlowControlIn(flowChoiceIn.getSelectedItem());
            parameters.setFlowControlOut(flowChoiceOut.getSelectedItem());
            parameters.setDatabits(databitsChoice.getSelectedItem());
            parameters.setStopbits(stopbitsChoice.getSelectedItem());
            parameters.setParity(parityChoice.getSelectedItem());
            
        }
    */
    }

    /**
    Handles closing down system. Allows application to be closed with window
    close box.
    */
    class CloseHandler extends WindowAdapter {

	AirboardFrame sd;
	
	public CloseHandler(AirboardFrame sd) {
	    this.sd = sd;
	}

	public void windowClosing(WindowEvent e) {
	    sd.shutdown();
	}
    }
}
