/*
 * AIRboardUI: graphical interface program to interact with airboard
 * through RS232 serial port.
 *
 * In order to compile this package you have to install the
 * Java Communications API on your system.
 * http://java.sun.com/products/javacomm/downloads/index.html
 * 
 * Bruschieri Stefano
 * Campari PierPaolo
 */
package airboardUI.menu;

import java.awt.Button;
import java.awt.Dimension;
import java.awt.FileDialog;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Label;
import java.awt.Menu;
import java.awt.MenuItem;
import java.awt.Panel;
import java.awt.TextField;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Properties;
import airboardUI.AlertDialog;
import airboardUI.AirboardFrame;
import airboardUI.SerialParameters;
import airboardUI.SerialParametersForm;
import java.util.Vector;
import javax.swing.JOptionPane;

/**
 * @author Bruschieri Stefano
 */
public class FileMenu extends Menu implements ActionListener{

	private AirboardFrame parent;
    private Properties props = null;
    private SerialParameters parameters;
	
	private MenuItem openItem;
	private MenuItem saveItem;
	private MenuItem exitItem;
	private MenuItem settingsItem;
	private MenuItem gsItem;
	
	public FileMenu(AirboardFrame parent) {
		super("File");
		this.parent = parent;
		this.parameters = this.parent.getParameters();
		
		openItem = new MenuItem("Load");
		openItem.addActionListener(this);
		this.add(openItem);

		saveItem = new MenuItem("Save");
		saveItem.addActionListener(this);
		this.add(saveItem);

		exitItem = new MenuItem("Exit");
		exitItem.addActionListener(this);
		this.add(exitItem);
		
		this.addSeparator();
		settingsItem = new MenuItem("Serial settings");
		settingsItem.addActionListener(this);
		this.add(settingsItem);
		
		gsItem = new MenuItem("General settings");
		gsItem.addActionListener(this);
		this.add(gsItem);
	}
	
    /**
    Responds to the menu items and buttons.
    */
    public void actionPerformed(ActionEvent e) {
    	String cmd = e.getActionCommand();
    	
    	// Loads a configuration file.
    	if (cmd.equals("Load")) {
    	    if (parent.getConnection().isOpen()) {
    		AlertDialog ad = new AlertDialog(this.parent, "Port Open!",
    						"Configuration may not"+
    						"be loaded"+
    						"while a port is open.");
    	    } else {
    		FileDialog fd = new FileDialog(this.parent, 
    					       "Load Port Configuration",
    					       FileDialog.LOAD);
    		fd.setVisible(true);
    		String file = fd.getFile();
    		if (file != null) {
    		    String dir = fd.getDirectory();
    		    File f = new File(dir + file);
    		    try {
    		    	FileInputStream fis = new FileInputStream(f);
    		    	props = new Properties();
    		    	props.load(fis);
    		    	fis.close();
    		    } catch (FileNotFoundException e1) {
    			System.err.println(e1);
    		    } catch (IOException e2) {
    			System.err.println(e2);
    		    }
    		    loadParams();
    	 	}
    	    }
    	}
    	
    	// Saves a configuration file.
    	if (cmd.equals("Save")) {
    	    FileDialog fd = new FileDialog(this.parent, "Save Port Configuration",
    						   FileDialog.SAVE);
    	    fd.setFile("AIRboardUI.properties");
    	    fd.setVisible(true);
    	    String fileName = fd.getFile();
    	    String directory = fd.getDirectory();
    	    if ((fileName != null) && (directory != null)) {
                    writeFile(directory + fileName);
                } 
    	}

    	// Calls shutdown, which exits the program.
    	if (cmd.equals("Exit")) {
    	    this.parent.shutdown();
    	}
    	
    	//Opens a GUI to change serial parameters
    	if(cmd.equals("Serial settings")){
    		SerialParametersForm spf = new SerialParametersForm(this.parameters, 
    				this.parent.getConnection(), 
					this.parent);
    	}
    	
    	//Opens a GUI to change general settings
    	if(cmd.equals("General settings")){
    		GeneralSettings gs = new GeneralSettings(this.parent);
    	}
    }

    /**
    Writes the current parameters to a configuration file of the 
    java.properties style.
    */
    private void writeFile(String path) {

        Properties newProps;
        FileOutputStream fileOut = null;

        newProps = new Properties();

        newProps.put("portName", parameters.getPortName());
        newProps.put("baudRate", parameters.getBaudRateString());
        newProps.put("flowControlIn", parameters.getFlowControlInString());
        newProps.put("flowControlOut", parameters.getFlowControlOutString());
        newProps.put("parity", parameters.getParityString());
        newProps.put("databits", parameters.getDatabitsString());
        newProps.put("stopbits", parameters.getStopbitsString());

        try {
            fileOut = new FileOutputStream(path);
        } catch (IOException e) {
            System.out.println("Could not open file for writiing");
        }

        newProps.save(fileOut, "AIRboardUI poperties");

        try {
            fileOut.close();
        } catch (IOException e) {
            System.out.println("Could not close file for writiing");
        }
    }
    
    /**
    Set the parameters object to the settings in the properties object.
    */
    private void loadParams() {
    	parameters.setPortName(props.getProperty("portName"));
    	parameters.setBaudRate(props.getProperty("baudRate"));
    	parameters.setFlowControlIn(props.getProperty("flowControlIn"));
    	parameters.setFlowControlOut(props.getProperty("flowControlOut"));
    	parameters.setParity(props.getProperty("parity"));
    	parameters.setDatabits(props.getProperty("databits"));
    	parameters.setStopbits(props.getProperty("stopbits"));
	
    	// Show to the user the updated serial parameters
    	this.parent.updateSettingsPanel();
    }
	
    
    /**
     * Create a form to insert some general settings about
     * the whose program
     */
    private class GeneralSettings extends Frame implements ActionListener{
    	
    	private static final int WIDTH = 400;
    	private static final int HEIGHT = 200;
    	
    	private Frame parent;
    	private Panel formPanel;
    	private Panel buttonPanel;
    	private Button applyButton;
    	private Button cancelButton;
    	private Button changeOneButton;
    	private Button changeTwoButton;
    	private Button changeRuleButton;
    	private Button createScriptButton;
        private TextField inputOneTextField;
        private TextField inputTwoTextField;
        private TextField inputRulesField;
        private TextField inputDelayTimeField;
    	
    	/* Write the minicom script to send data through serial port */
        private FileOutputStream out = null; 
        
    	public GeneralSettings(Frame parent){
    		super("General settings");
    		this.parent = parent;
    		    		  		
    		formPanel = new Panel();
    		applyButton = new Button("Apply");
    		applyButton.addActionListener(this);
    		cancelButton = new Button("Cancel");
    		cancelButton.addActionListener(this);
    		changeOneButton = new Button("Change One");
    		changeOneButton.addActionListener(this);
    		changeTwoButton = new Button("Change Two");
    		changeTwoButton.addActionListener(this);
    		changeRuleButton = new Button("Change Rule");
    		changeRuleButton.addActionListener(this);
    		createScriptButton = new Button("Create");
    		createScriptButton.addActionListener(this);
    		inputOneTextField = new TextField(AirboardFrame.FIRST_INPUT_FILE, 20);
    		inputOneTextField.setEditable(true);
    		inputTwoTextField = new TextField(AirboardFrame.SECOND_INPUT_FILE, 20);
    		inputTwoTextField.setEditable(true);
    		inputRulesField = new TextField(AirboardFrame.RULES_FILE, 20);
    		inputRulesField.setEditable(true);
    		inputDelayTimeField = new TextField(String.valueOf(AirboardFrame.DELAY_TIME), 20);
    		inputDelayTimeField.setEditable(true);
    		
    		GridBagLayout gridBag = new GridBagLayout();
    		GridBagConstraints cons = new GridBagConstraints();
    		cons.gridwidth = GridBagConstraints.REMAINDER;
    		cons.weightx = 1.0;
    		formPanel.setLayout(new GridLayout(6,3));
    		formPanel.add(new Label("First input file:"));
    		formPanel.add(inputOneTextField);
    		formPanel.add(changeOneButton);
    		formPanel.add(new Label("Second input file:"));
    		formPanel.add(inputTwoTextField);
    		formPanel.add(changeTwoButton);
    		formPanel.add(new Label("Rules input file:"));
    		formPanel.add(inputRulesField);
    		formPanel.add(changeRuleButton);
    		formPanel.add(new Label("Delay time:"));
    		formPanel.add(inputDelayTimeField);
    		formPanel.add(new Label(""));
    		formPanel.add(new Label("Create minicom script:"));
    		formPanel.add(new Label(""));
    		formPanel.add(createScriptButton);
    		gridBag.setConstraints(formPanel, cons);
    		add(formPanel);

    		buttonPanel = new Panel();
    		buttonPanel.add(applyButton);
    		buttonPanel.add(cancelButton);
    		//buttonPanel.add(changeButton);
    		add(buttonPanel, "Center");

    		this.setLayout(gridBag);
    		
    		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
    		this.setLocation(screenSize.width/2 - WIDTH/2, 
    			    screenSize.height/2 - HEIGHT/2);
    		this.setSize(WIDTH, HEIGHT);
    		this.setVisible(true);
    		this.repaint();	
    	}
    	
        /**
        Responds to the menu items and buttons.
        */
        public void actionPerformed(ActionEvent e) {
        	String cmd = e.getActionCommand(); 
        	
        	// Close the settings window
        	if (cmd.equals("Cancel")) {
        	    this.setVisible(false);
        	}
        	
        	// Change the first input file
        	if (cmd.equals("Change One")) {
        	    FileDialog fd = new FileDialog(this, "Save Port Configuration",
						   FileDialog.LOAD);
        	    fd.setVisible(true);
        		String file = fd.getFile();
        		if (file != null /*contains .dat*/) {
        		    String dir = fd.getDirectory();
        		    System.out.println( dir+file );
        		    dir = dir + file;
        		    inputOneTextField.setText(dir);	    
        	 	}
        	}
        	
        	// Change the second input file
        	if (cmd.equals("Change Two")) {
        	    FileDialog fd = new FileDialog(this, "Save Port Configuration",
						   FileDialog.LOAD);
        	    fd.setVisible(true);
        		String file = fd.getFile();
        		if (file != null /*contains .dat*/) {
        		    String dir = fd.getDirectory();
        		    System.out.println( dir+file );
        		    dir = dir + file;
        		    inputTwoTextField.setText(dir);	    
        	 	}
        	}
        	
        	// Change the rules file
        	if (cmd.equals("Change Rule")) {
        	    FileDialog fd = new FileDialog(this, "Save Port Configuration",
						   FileDialog.LOAD);
        	    fd.setVisible(true);
        		String file = fd.getFile();
        		if (file != null /*contains .dat*/) {
        		    String dir = fd.getDirectory();
        		    System.out.println( dir+file );
        		    dir = dir + file;
        		    inputRulesField.setText(dir);	    
        	 	}
        	}
        	
        	// Apply changes
        	if (cmd.equals("Apply")) {
        		// Texts in fields must be different from null value
        		if( inputOneTextField.getText() != null ){
        			AirboardFrame.FIRST_INPUT_FILE = inputRulesField.getText();
        		}
        		if( inputTwoTextField.getText() != null ){
        			AirboardFrame.SECOND_INPUT_FILE = inputRulesField.getText();
        		}
        		if( inputRulesField.getText() != null ){
        			AirboardFrame.RULES_FILE = inputRulesField.getText();
        		}
        		if( inputDelayTimeField.getText() != null){
            		int newDelayTime = (new Integer(inputDelayTimeField.getText())).intValue();
        			AirboardFrame.DELAY_TIME = newDelayTime;
        		}
        		this.setVisible(false);
        	}
        	
        	// Create the minicom script to send fuzzy table
        	// through serial connection
        	if (cmd.equals("Create")) {
        		// make three vectors to load data from files
        		Vector firstInputVector = null;
        		Vector secondInputVector = null;
        		Vector rulesVector = null;
        		String script = "";
        		try {
        			firstInputVector = ((AirboardFrame)parent).loadFuzzyData(AirboardFrame.FIRST_INPUT_FILE);
        			secondInputVector = ((AirboardFrame)parent).loadFuzzyData(AirboardFrame.SECOND_INPUT_FILE);
        			rulesVector = ((AirboardFrame)parent).loadFuzzyData(AirboardFrame.RULES_FILE);
        		} catch (IOException ioe) {
        			System.err.println("An error was occurred "+ioe);
        		}
        		script = "send d2\n";
        		
        		for(int i=0; i<firstInputVector.size(); i++){
        			String tempString = "", tempIndex = "";
        			tempString = (firstInputVector.elementAt(i)).toString();
        			tempIndex = String.valueOf(i);
    		    	if(tempString.length()==1) tempString = "00"+tempString;
    		    	if(tempString.length()==2) tempString = "0"+tempString;
    		    	if(tempIndex.length()==1) tempIndex = "00"+tempIndex;
    		    	if(tempIndex.length()==2) tempIndex = "0"+tempIndex;
    		    	script = script + "send ewfuzzymf1=+"+tempString+"&+"+tempIndex+"\n"+
					"expect {\n"+
					"\"OK\\r\" break\n"+
					"timeout 3 print \"NO RESPONSE or ERROR writing ewfuzzymf1=+"+tempString+"&+"+tempIndex+"\"\n"+
					"}\n"+
					"print WritingNext\n";
        		}

        		for(int j=0; j<secondInputVector.size(); j++){
        			String tempString = "", tempIndex = "";
        			tempString = (secondInputVector.elementAt(j)).toString();
        			tempIndex = String.valueOf(j);
        			//script = script + "send ";
    		    	if(tempString.length()==1) tempString = "00"+tempString;
    		    	if(tempString.length()==2) tempString = "0"+tempString;
    		    	if(tempIndex.length()==1) tempIndex = "00"+tempIndex;
    		    	if(tempIndex.length()==2) tempIndex = "0"+tempIndex;
    		    	script = script + "send ewfuzzymf2=+"+tempString+"&+"+tempIndex+"\n"+
					"expect {\n"+
					"\"OK\r\" break\n"+
					"timeout 3 print \"NO RESPONSE or ERROR writing ewfuzzymf2=+"+tempString+"&+"+tempIndex+"\"\n"+
					"}\n"+
					"print WritingNext\n";
        		}
        		
        		// write number of rules before sendings each rule
        		String ruleNum = String.valueOf(rulesVector.size());
		    	if(ruleNum.length()==1) ruleNum = "00"+ruleNum;
		    	if(ruleNum.length()==2) ruleNum = "0"+ruleNum;
		    	script = script +"ewfrulenum=+"+ruleNum+"&+000\n";

        		for(int h=0; h<rulesVector.size(); h++){
        			String tempString = "", tempIndex = "";
        			tempString = (rulesVector.elementAt(h)).toString();
        			tempIndex = String.valueOf(h);
    		    	if(tempString.length()==1) tempString = "00"+tempString;
    		    	if(tempString.length()==2) tempString = "0"+tempString;
    		    	if(tempIndex.length()==1) tempIndex = "00"+tempIndex;
    		    	if(tempIndex.length()==2) tempIndex = "0"+tempIndex;
    		    	script = script + "send ewfuzrules=+"+tempString+"&+"+tempIndex+"\n"+
					"expect {\n"+
					"\"OK\r\" break\n"+
					"timeout 3 print \"NO RESPONSE or ERROR writing ewfuzrules=+"+tempString+"&+"+tempIndex+"\"\n"+
					"}\n"+
					"print WritingNext\n";
        		}
        		
            	File outputFile = new File("airboardUI/data/airboard_script");
            	
            	try{
            		this.out = new FileOutputStream(outputFile);
                	for( int k=0; k<script.length(); k++)
                		this.out.write((int)script.charAt(k));
                	this.out.close();
                	JOptionPane.showMessageDialog(this, "Creating minicom script ends succesfully", 
                			"Writing ok", JOptionPane.INFORMATION_MESSAGE);
            	} catch (Exception ferror){
        			AlertDialog np = new AlertDialog(this, "Error on file",
        					"Error opening or writing minicom script"+
							ferror);
            		System.out.println("Error opening file "+ferror);
            	}
        	}
        }

    }
}
