
#include "VisionExpert.h"
#include <vector>
#include <boost/program_options.hpp>
#include "MatrixFunctions.h"
#include "msg.h"
#include <MrtLogger.h>
#include <WheelChairUtil.h>

using namespace std;

bool VisionExpert::isReady(){
	return (camera!=NULL);
}

int VisionExpert::initCoordAssPostFromFile(){
	double 	x;
	double 	y;
	double 	z;
	double 	a;
	double 	b;
	double 	c;
	
	using namespace po;
	options_description generic;
	variables_map vm;	
	try {
		generic.add_options()
		    ("x", 		value<double >(&x), 					"x")
		    ("y", 		value<double >(&y), 					"y")
		    ("z", 		value<double >(&z), 					"z")
		    ("alpha", 		value<double >(&a), 					"a")
		    ("beta", 		value<double >(&b), 					"b")
		    ("gamma", 		value<double >(&c), 					"g")
		    ;
	       
		ifstream ifs(GestioneConfigurazione::getAbsoluteCoordinateWheelchairToCameraRelationFile().c_str());
		store(parse_config_file(ifs, generic), vm);
		notify(vm);
	}
	catch (exception& e){
		cout << e.what() << "\n";
		return -1;
	}        
	coordAss->setPostMatrix(x,y,z,a,b,c);
	DBG("vision:: coord ass post mult added %d\n");
	return 0;
}

int VisionExpert::initCoordAssFromFile(){
	vector<int> 	tagNum;
	vector<double> 	x;
	vector<double> 	y;
	vector<double> 	z;
	vector<double> 	a;
	vector<double> 	b;
	vector<double> 	c;
	
	using namespace po;
	options_description generic;
	variables_map vm;	
	try {
		generic.add_options()
		    ("tagNumber", 	value<vector<int> >(&tagNum), 					"list 	of knew tags")
		    ("tagX", 		value<vector<double> >(&x), 					"x 	of knew tags")
		    ("tagY", 		value<vector<double> >(&y), 					"y 	of knew tags")
		    ("tagZ", 		value<vector<double> >(&z), 					"z 	of knew tags")
		    ("tagAlpha", 	value<vector<double> >(&a), 					"alpha 	of knew tags")
		    ("tagBeta", 	value<vector<double> >(&b), 					"beta 	of knew tags")
		    ("tagGamma", 	value<vector<double> >(&c), 					"gamma 	of knew tags")
		    ;
	       
		ifstream ifs(GestioneConfigurazione::getAbsoluteCoordinateTagRelationFile().c_str());
		store(parse_config_file(ifs, generic), vm);
		notify(vm);
	}
	catch (exception& e){
		cout << e.what() << "\n";
		return -1;
	}        
	if(tagNum.size()!=x.size())return -1;
	if(tagNum.size()!=y.size())return -1;
	if(tagNum.size()!=z.size())return -1;
	if(tagNum.size()!=a.size())return -1;
	if(tagNum.size()!=b.size())return -1;
	if(tagNum.size()!=c.size())return -1;
	for(int i=0;i<tagNum.size();i++){
		coordAss->addTag(tagNum[i],x[i],y[i],z[i],a[i],b[i],c[i]);
		DBG("vision:: coord ass added %d\n",tagNum[i]);
	}

return 0;
}

int VisionExpert::initInvMatFromFile(){
	vector<int> 	tagNum;
	vector<double> 	zeta;
	vector<double> 	beta;
	vector<double> 	gamma;

	using namespace po;
	options_description generic;
	variables_map vm;	
	try {
		generic.add_options()
		    ("tagNumber", 	value<vector<int> >(&tagNum), 					"list of knew tags")
		    ("tagZeta", 	value<vector<double> >(&zeta), 					"zeta of knew tags")
		    ("tagBeta", 	value<vector<double> >(&beta), 					"beta of knew tags")
		    ("tagGamma", 	value<vector<double> >(&gamma), 				"gamma of knew tags")
		    ;
	       
		ifstream ifs(GestioneConfigurazione::getMatrixInversionParameterFile().c_str());
		store(parse_config_file(ifs, generic), vm);
		notify(vm);
	}
	catch (exception& e){
		cout << e.what() << "\n";
		return -1;
	}        
	if(tagNum.size()!=zeta.size())return -1;
	if(tagNum.size()!=beta.size())return -1;
	if(tagNum.size()!=gamma.size())return -1;
	for(int i=0;i<tagNum.size();i++){
		invMat->addTag(tagNum[i],zeta[i],beta[i],gamma[i]);
		DBG("vision:: inv correction added %d\n",tagNum[i]);
	}
	return 0;
}

VisionExpert::VisionExpert(ModuleDispatch* kernel,int period,bool simFileCamera,string prefix,string postfix,string format) : StringModuleMember(kernel,period,"vision"){
	if(!simFileCamera){
		camera= new FireIAcquisition(GestioneConfigurazione::getFireIdevice(),GestioneConfigurazione::getSaveFrame(),GestioneConfigurazione::getSaveImageNamePrefix(),GestioneConfigurazione::getSaveImageNameSubfix(),GestioneConfigurazione::getSaveImageNameFormat());
		MrtLogger::writeLog("Vision Expert :: init with camera");
	}
	else{
		camera=new FileRawAcquisition(prefix,postfix,format);
		MrtLogger::writeLog("Vision Expert :: simulation with prefix: '"+prefix+"' postfix: '"+postfix+"' format: '"+format+"'");
	}
	if(!camera->isReady()){
		MrtLogger::writeLog("Vision Expert :: error init camera");
		delete camera;
		camera=NULL;
	}
	art= new ARToolKitWrapper(GestioneConfigurazione::getArtoolkitCalibFile(),GestioneConfigurazione::getArtoolkitTagDimension());

	invMat= new InverseMatrixCorrection();
	if(initInvMatFromFile()!=0){
		fprintf(stderr,"vision::error in inverse matrix init\n");
	}
	
	coordAss= new GestioneCoordinateAssolute();
	if(initCoordAssFromFile()!=0){
		fprintf(stderr,"vision::error in absolute coordinate init\n");
	}
	if(initCoordAssPostFromFile()!=0){
		fprintf(stderr,"vision::error in absolute coordinate post init\n");
	}
	
}

VisionExpert::~VisionExpert (){
	if(camera!=NULL)delete camera;
	delete invMat;
	delete art;
	delete coordAss;
}

void VisionExpert::RunInit(){
	MrtLogger::writeLog("Vision Expert run init");
	if(camera!=NULL)camera->Start();
	struct timezone tz;
	gettimeofday(&tstart,&tz);//uso tstart per non mandare l'intero time stamp, ma un delta rispetto l'inizio
	
	/*NewMessage(MSG_POSITION,POS_DATA);
	mCurrentMessage<<" posx "<<STARTX<<" posy "<<STARTY<<" angle "<<STARTANGLE<<" \n";
	CloseMessage();
	SendAllStringMessages();*/
	
	for(int i=0;i<POSITIONFILTER;i++){
		posx_array[i].v=0;
		posy_array[i].v=0;
		posa_array[i].v=0;
		posx_array[i].w=0;
		posy_array[i].w=0;
		posa_array[i].w=0;
	}
	
}

void VisionExpert::RunClose(){
	MrtLogger::writeLog("Vision Expert run close");
	if(camera!=NULL)camera->Stop();
}

void VisionExpert::RunDuty(){
	struct timeval 	tstamp,tdiff;
	struct timezone tz;
	char s[256];
	
	int flag=0,flag2=0;
	int notnull;
	int posvalid;
	float posx,posy,posa;
	weightfloat nposx,nposy,nposa;
	weightfloat newdata;
	
	float sx,sy,sa;
	float radiuscorrection;
	float anglecorrection;

	
	MrtLogger::writeLog("Vision Expert run duty");
	if(camera==NULL){
		MrtLogger::writeLog("Vision Expert :: no camera");	
		printf("no camera!\n");
		return;
	}
	unsigned char * img=camera->CaptureFrame();

	gettimeofday(&tstamp,&tz);
	timersub(&tstamp,&tstart,&tdiff);
	
	if(img!=NULL){
		int mark=art->processImage(img);


		DBG("vision:: marker %d\n",mark);
		if(mark>=0){
			ostringstream ss;
			ss<<"Vision Expert image marker info::\t"<<mark<<"\t"<<camera->getCount()<<"\t";
			//printf("expert: found",markerId);

			for(int i=0;i<4;i++){
				//cout<<" x :"<<art->getXPoint()[i] <<" y:"<<art->getXPoint()[i]<<"\n";
				ss<<art->getXPoint()[i]<<"\t";
				ss<<art->getYPoint()[i]<<"\t";
			}
			MrtLogger::writeLog(ss.str());

			if(invMat->getInverseMatrix(art->getMatrix3D(),mark)==0){
				DBG("vision:: inversion done\n",mark);
				//print3DMatrix(invMat->getResult(),"vision::inversion\n","\n\n");
				if(coordAss->getAbsolute(mark,invMat->getResult())==0){
					DBG("vision:: absolute coord calculated\n",mark);
					//ok: ho le coordinate della carrozzina assolute! (basta estrarre x,y,a)
					//inviarle al "gestore delle posizioni"
					
					/*NewMessage(MSG_FROM_VISION_POS,VISION_POS);
					mCurrentMessage <<D_OPEN_DATUM <<VISION_MARKER_ID	<< " " << mark 			<<D_CLOSE_DATUM;  
					sprintf(s,"%d",tdiff.tv_sec);
					mCurrentMessage <<D_OPEN_DATUM <<VISION_TIME_STAMP_SEC	<< " " << s 			<<D_CLOSE_DATUM;  
					sprintf(s,"%0.6d",tdiff.tv_usec);
					mCurrentMessage <<D_OPEN_DATUM <<VISION_TIME_STAMP_USEC	<< " " << s 			<<D_CLOSE_DATUM;  
					mCurrentMessage.precision(10);
					mCurrentMessage <<fixed<<D_OPEN_DATUM <<VISION_X		<< " " << coordAss->getX() 	<<D_CLOSE_DATUM;  
					mCurrentMessage <<D_OPEN_DATUM <<VISION_Y		<< " " << coordAss->getY() 	<<D_CLOSE_DATUM;  
					mCurrentMessage <<D_OPEN_DATUM <<VISION_ANGLE		<< " " << coordAss->getA() 	<<D_CLOSE_DATUM;  
					mCurrentMessage <<endl;
					SendAllStringMessages();*/
					
					posx=coordAss->getX();
					posy=coordAss->getY();
					posa=coordAss->getA();
					
					newdata.w=1;
					newdata.v=posx;
					nposx=getFilteredData(newdata,posx_array,POSITIONFILTER,POSITIONATT,&notnull);
					newdata.v=posy;
					nposy=getFilteredData(newdata,posy_array,POSITIONFILTER,POSITIONATT,&notnull);
					newdata.v=posa;
					nposa=getFilteredData(newdata,posa_array,POSITIONFILTER,POSITIONATT,&notnull);
					
					flag=1;
					
					ostringstream s;
					s<<"Vision Expert actual position::\t"<<coordAss->getX()<<"\t"<<coordAss->getY()<<"\t"<<coordAss->getA();
					MrtLogger::writeLog(s.str(),&tstamp);	
					//cout <<s.str()<<endl;	
				}
				else{
					MrtLogger::writeLog("Vision Expert actual position:: unknow, no marker id correspondance in absolute coordinate system");
					//printf("vision:: no marker id recognized\n");
				}
			}
			else{
				MrtLogger::writeLog("Vision Expert actual position:: unknow, no marker id correspondance in inversion");
				//printf("vision:: marker %d not recognized\n",mark);			
			}
		}
		else{
			MrtLogger::writeLog("Vision Expert actual position:: unknow, there are no marker recognized");	
			//printf("vision:: no marker recognized\n");
		}
	}
	else {
		MrtLogger::writeLog("Vision Expert actual position:: unknow, no image");	
		//printf("vision:: no image\n");		
	}
	
	if(flag==0){
		newdata.w=0;
		nposx=getFilteredData(newdata,posx_array,POSITIONFILTER,POSITIONATT,&notnull);
		nposy=getFilteredData(newdata,posy_array,POSITIONFILTER,POSITIONATT,&notnull);
		nposa=getFilteredData(newdata,posa_array,POSITIONFILTER,POSITIONATT,&notnull);
		//printf("\n%d\n",notnull);
		if(!(notnull<POSITIONFILTER)){
			flag2=1;
		}
	}
	
	if(flag2==0){
		sx=nposx.v/VISIONSCALEX + VISIONTRAX;
		sy=nposy.v/VISIONSCALEY +VISIONTRAY;
		sa=norangle(radtodeg(coordAss->getA())+180);
		radiuscorrection=52;
		anglecorrection=degtorad(-20);
		sx += cos(coordAss->getA()+anglecorrection)*radiuscorrection;
		sy += sin(coordAss->getA()+anglecorrection)*radiuscorrection;
		#ifndef ODOVISIONPOSITION
		NewMessage(MSG_POSITION,POS_DATA);
		#else
		NewMessage(MSG_POSITION_P,POS_DATA);
		#endif
		mCurrentMessage
		<<" posx "<<
		sx
		<<" posy "<<
		sy
		<<" angle "<<
		sa
		<<" \n";
		CloseMessage();
		SendAllStringMessages();
	}else{
		#ifndef ODOVISIONPOSITION
		NewMessage(MSG_POSITION,POS_DATA);
		#else
		NewMessage(MSG_POSITION_P,POS_DATA);
		#endif
		mCurrentMessage<<" invalidposition \n";
		CloseMessage();
		SendAllStringMessages();
	}		
	
					/*	NewMessage(MSG_POSITION,POS_DATA);
					mCurrentMessage
					<<" posx "<<
					((nposx.v/VISIONSCALEX + VISIONTRAX) + (cos((coordAss->getA())))*60)				
					<<" posy "<<
					((nposy.v/VISIONSCALEY +VISIONTRAY)  + (sin((coordAss->getA())))*60)
					<<" angle "<<norangle(radtodeg(coordAss->getA())+180)<<" \n";
					CloseMessage();
					SendAllStringMessages();*/
		
}

